/** 
 *   5850 Absolute shaft encoder driver.
 *
 *   (c) Copyright 2003 Federico Bareilles <fede@iar.unlp.edu.ar>,
 *   Instituto Argentino de Radio Astronomia (IAR).
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *     
 *   The author does NOT admit liability nor provide warranty for any
 *   of this software. This material is provided "AS-IS" in the hope
 *   that it may be useful for others.
 *
 *   TODO: 
 *      *) spin_lock.
 *      -) data comunication.
 *      *) if load fail: exit is not clean.
 *      -) file_operations: read 
 *      -) file_operations: ioctl
 *      *)
 *      *)
 *      *)
 *      *)
 *
 **/

#define E5850_VERSION "0.01"

#define USING_PROC_FS

#ifdef USING_PROC_FS
#define ATOMIC_VAL
#endif

#include <linux/module.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <linux/reboot.h>
#include <linux/init.h>
#include <linux/proc_fs.h>
#include <linux/ioport.h>
#include <linux/delay.h>  /* udelay */
#include "e5850.h"

#ifndef CONFIG_PROC_FS
#undef USING_PROC_FS
#undef ATOMIC_VAL
#endif

#define OUR_NAME "encoder5850"

/* Default bits resolution (16 bits maximum) */
#define E5850_NB 14

/* Port definitions: */
#define BASE_IO 0x378

/**
 * The pinout of the parelllel port for absolute encoder 5850 and
 * anemometer
 *
 *      Pin  |  Signal                | Bit  |  Port
 *      -----+------------------------+------+-----------------
 *        2  |  Pulse for 5850's      |  0   |  Data   (base+0)
 *       10  |  Siderial clock (IRQ)  |  6   |  Status (base+1)
 *       12  |  Dara read encoder 1   |  5   |  Status (base+1)
 *       13  |  Data read encoder 0   |  4   |  Status (base+1)
 *       15  |  Data read anemometer 0|  3   |  Status (base+1)
 **/

/* Pulse use bits 0 of data port (pin 2) */
#define PULSE_UP   0x01
#define PULSE_DOWN 0x00

/* firts encoder bits positions (pin 13) */
#define MASK_0 0x10 
#define OFFSET_0 4
/* second encoder bits positions (pin 12) */
#define MASK_1 0x20 
#define OFFSET_1 5
/* anemometer bits positions (pin 15) */
#define MASK_2 0x08 
#define OFFSET_2 3


static int major = 0; /* dynamic by default */
MODULE_PARM(major, "i");
MODULE_PARM_DESC(major,"Major number for /dev/e5850; dynamic by default.");

/* default is the first printer port on PC's. "e5850_base" is there
   too because it's what we want to use in the code */
static unsigned long base = BASE_IO;
unsigned long e5850_base = 0;
MODULE_PARM(base, "i");
MODULE_PARM_DESC(base,"Base io port (default=0x378)");

/* The interrupt line is undefined by default. "e5850_irq" is as above */
static int irq = -1;
volatile int e5850_irq = -1;
MODULE_PARM(irq, "i");
MODULE_PARM_DESC(irq,"irq line");

static int nbits = E5850_NB;
volatile int e5850_nb;
MODULE_PARM(nbits, "i");
MODULE_PARM_DESC(nbits,"Number of bits resolution");


#ifdef USING_PROC_FS
struct proc_dir_entry *e5850_proc;
#ifdef ATOMIC_VAL
atomic_t encoder_0;
atomic_t encoder_1;
atomic_t encoder_2;
#endif
#endif

static unsigned long e5850_is_open;
static unsigned long e5850_is_running;

/* definitions for cicular bufer */
static unsigned long e5850_buffer = 0;
volatile unsigned long e5850_head;
volatile unsigned long e5850_tail;

static wait_queue_head_t e5850_queue; 
//DECLARE_WAIT_QUEUE_HEAD(e5850_queue);

/*
 * Atomicly increment an index into e5850_buffer
 */
static inline void e5850_incr_bp(volatile unsigned long *index, int delta)
{
	unsigned long new = *index + delta;
	barrier ();  /* Don't optimize these two together */
	*index = (new >= (e5850_buffer + PAGE_SIZE)) ? e5850_buffer : new;
}


void e5850_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
	int i;
	int v, val0 = 0, val1 = 0, val2 = 0;
	int written;
	
	/* first triger */
	outb_p(PULSE_DOWN, e5850_base);
	/* main loop */
	for(i = e5850_nb - 1; i >=  0; i--){
		outb_p(PULSE_UP, e5850_base);
		v = inb_p(e5850_base+1);
		val0 |= ((v & MASK_0) >> (OFFSET_0) ) << i;
		val1 |= ((v & MASK_1) >> (OFFSET_1) ) << i;
		val2 |= ((v & MASK_2) >> (OFFSET_2) ) << i;
		outb_p(PULSE_DOWN, e5850_base);
	}
#if 1
	if(test_bit(0, &e5850_is_running)){
		/* Write a 16 byte record. Assume PAGE_SIZE is a
                   multiple of 16 */
		written = sprintf((char *)e5850_head," %04x %04x %04x\n",
				  val0, val1, val2);
		e5850_incr_bp(&e5850_head, /*written*/ 16);
		/* awake any reading process */
//		wake_up_interruptible(&e5850_queue);
	}
#endif

#ifdef ATOMIC_VAL
	atomic_set( &encoder_0, val0);
	atomic_set( &encoder_1, val1);
	atomic_set( &encoder_2, val2);
#endif
	outb_p(PULSE_UP, e5850_base);
}


#ifdef USING_PROC_FS
#ifdef ATOMIC_VAL
static int e5850_atomic(char *buffer, char **start, off_t fpos, int length)
{
        char *temp;
	
        temp = buffer;
        temp += sprintf(temp, "Encoder0: 0x%04x\n", atomic_read(&encoder_0));
        temp += sprintf(temp, "Encoder1: 0x%04x\n", atomic_read(&encoder_1));
        temp += sprintf(temp, "Encoder2: 0x%04x\n", atomic_read(&encoder_2));

        return temp-buffer;
}
#endif

static int e5850_info(char *buffer, char **start, off_t fpos, int length)
{
        char *temp;
	
        temp = buffer;
        temp += sprintf(temp, "Count:   %ld\n", e5850_head - e5850_tail);
        temp += sprintf(temp, "Open:    %1d\n",test_bit(0, &e5850_is_open));
        temp += sprintf(temp, "Running: %1d\n",test_bit(0, &e5850_is_running));
	
        return temp-buffer;
}
#endif


/**
 *      e5850_open        -       handle open of e5850 encoder
 *      @inode: inode of device
 *      @file: file handle to device
 *
 *      Open the e5850 encoder device. Ensure only one person opens it
 *      at a time. Also start the circular buffer.
 */

static int e5850_open(struct inode *inode, struct file *file)
{
	if(test_and_set_bit(0, &e5850_is_open))
                return -EBUSY;
	set_bit(0, &e5850_is_running);

	return 0;
}

/**
 *      e5850_release     -       close 
 *      @inode: inode from VFS
 *      @file: file from VFS
 */

static int e5850_release(struct inode *inode, struct file *file)
{
	clear_bit(0, &e5850_is_running);
	clear_bit(0, &e5850_is_open);
	
	return 0;
}

ssize_t e5850_read(struct file *filp, char *buf, size_t count, loff_t *f_pos)
{
	int count0;
	
#if 0 /* FIXME */
/**
 *  In this case, the value of e585_head could change between the test
 *  in the while statement and the call to interruptible_sleep_on. In
 *  that case, the driver will sleep even though new data is
 *  available;
 **/
	while (e5850_head == e5850_tail) {
		interruptible_sleep_on(&e5850_queue);
		if (signal_pending (current))  /* a signal arrived */
			/* tell the fs layer to handle it */
			return -ERESTARTSYS; 
		/* else, loop */
	}
#else
	/* This macro was added during 2.1 development to make life
           easier: */
	wait_event_interruptible(e5850_queue,e5850_head == e5850_tail);
#endif
	/* count0 is the number of readable data bytes */
	count0 = e5850_head - e5850_tail;
	if (count0 < 0) /* wrapped */
		count0 = e5850_buffer + PAGE_SIZE - e5850_tail;
	if (count0 < count) count = count0;
	
	if (copy_to_user(buf, (char *)e5850_tail, count))
		return -EFAULT;
	e5850_incr_bp (&e5850_tail, count);
	return count;
}


/**
 *      e5850_ioctl       -       handle encoders ioctls
 *      @inode: inode of the device
 *      @file: file handle to the device
 *      @cmd: command
 *      @arg: argument pointer
 *
 *      Handle the encoders ioctls supported by the 5850 driver.
 **/

static int e5850_ioctl(struct inode *inode, struct file *file,
		       unsigned int cmd, unsigned long arg)
{
	int state;

	switch (cmd) {
	case E5850_GETBUFFLEN:
		return put_user((int) (e5850_head - e5850_tail),
				(int *) arg);
	case E5850_CLEARBUFF:
		state = test_and_clear_bit(0, &e5850_is_running);
		e5850_head = e5850_tail = 0;
		if(state) set_bit(0, &e5850_is_running);
		return 0;
	case E5850_START:
		set_bit(0, &e5850_is_running);
		return 0;
	case E5850_STOP:
		clear_bit(0, &e5850_is_running);
		return 0;
	case E5850_ISRUNNING:
		return test_bit(0, &e5850_is_running);
        default:
                return -ENOIOCTLCMD;
		
        }
}



static struct file_operations e5850_fops = {
	.owner          = THIS_MODULE,
	.read           = e5850_read,
	.ioctl          = e5850_ioctl,
	.open           = e5850_open,
	.release        = e5850_release,
};


static int __init e5850_init(void)
{
	int result;
	
	/**
	 * first, encoder5850 out the base/e5850_base ambiguity: we'd
	 * better use e5850_base in the code, for clarity, but allow
	 * setting just "base" at load time. Same for "irq" and
	 * "nbits".
	 **/
	e5850_base = base;
	e5850_irq = irq;
	e5850_nb = nbits;
	
	result = check_region(e5850_base, 2);
	if (result) {
		printk(KERN_INFO OUR_NAME 
		       ": can't get I/O port address 0x%lx\n",
		       e5850_base);
		return result;
	}
	request_region(e5850_base, 2, OUR_NAME);
	
	result = register_chrdev(major, "e5850", &e5850_fops);
	if (result < 0) {
		printk(KERN_INFO OUR_NAME
		       ": can't get major number\n");
		release_region(e5850_base,2);
		return result;
	}
	if (major == 0) major = result; /* dynamic */

	/* FIXME */
	init_waitqueue_head (&e5850_queue);

	e5850_buffer = __get_free_pages(GFP_KERNEL,0); /* never fails */
	if (! e5850_buffer) {
		printk(KERN_ERR OUR_NAME
		       ": cannot get free pages for e5850_buffer\n");
	}
	e5850_head = e5850_tail = e5850_buffer;
	
	/**
	 * Set default interupt.
	 **/
	if (e5850_irq < 0) /* not yet specified: force the default on */
		switch(e5850_base) {
		case 0x378: e5850_irq = 7; break;
		case 0x278: e5850_irq = 2; break;
		case 0x3bc: e5850_irq = 5; break;
		}
	
	if (e5850_irq >= 0) {
		result = request_irq(e5850_irq, e5850_interrupt,
				     SA_INTERRUPT, OUR_NAME, NULL);
		if (result) {
			printk(KERN_INFO OUR_NAME 
			       ": can't get assigned irq %i\n",
			       e5850_irq);
			e5850_irq = -1;
    }
		else { /* actually enable it -- assume this *is* a
                          parallel port */
			outb(0x10,e5850_base+2);
		}
	}

#ifdef USING_PROC_FS
	e5850_proc = proc_mkdir("e5850", 0);
	if (!e5850_proc) {
		printk (KERN_ERR "cannot init /proc/e5850\n");
		return -ENOMEM;
	}
	create_proc_info_entry("e5850/info", 0, NULL, e5850_info);
#ifdef ATOMIC_VAL
	atomic_set( &encoder_0, 0);
	atomic_set( &encoder_1, 0);
	atomic_set( &encoder_2, 0);
	create_proc_info_entry("e5850/atomic_val", 0, NULL, e5850_atomic);
#endif
#endif
	
	/* Up Pulse Line */
	outb_p(PULSE_UP, e5850_base);   
	
	printk(KERN_INFO "Absolute encoder 5850 v(" 
	       E5850_VERSION ") (%d bits) initialising at io 0x%lx irq %d.\n",
	       e5850_nb, e5850_base, e5850_irq);
	
	return 0;
}


static void __exit e5850_exit(void)
{
	if (e5850_irq >= 0) {
		outb(0x0, e5850_base + 2);   /* disable the interrupt */
		free_irq(e5850_irq, NULL);
	}
	/* ! Down Pulse Line */
	outb_p(PULSE_UP, e5850_base);
	unregister_chrdev(major, "e5850");
	release_region(e5850_base, 2);
#ifdef USING_PROC_FS
	remove_proc_entry("e5850/info",NULL);
#ifdef ATOMIC_VAL
	remove_proc_entry("e5850/atomic_val",NULL);
#endif
	remove_proc_entry ("e5850", 0);
#endif
	if (e5850_buffer) free_page(e5850_buffer);

	printk(KERN_INFO "Absolute encoder 5850 v(" 
	       E5850_VERSION ") down.\n");
	return;
}

module_init(e5850_init);
module_exit(e5850_exit);

MODULE_AUTHOR("Federico Bareilles <fede@iar.unlp.edu.ar>");
MODULE_DESCRIPTION("5850 Absolute encoder");
//MODULE_LICENSE("GPL");
/* MODULE_SUPPORTED_DEVICE("dev"); */

EXPORT_NO_SYMBOLS;
